@@ -38,6 +38,8 @@ class Metrc:
         Facilities that are identified as dispensaries.
     mipps : list[str]
         Facilities that are identified for manufacturing and processing.
+    providers : list[str]
+        Facilities that are identified as cultivators.T
     '''
 
     def __init__(self, vendor_api_key, user_api_key):
@@ -57,6 +59,30 @@ class Metrc:
         self.api = APICaller(self.vendor_api_key, self.user_api_key)
         self.init_facilities()
 
+    def init_facilities(self):
+        '''
+        Get and store facility data from Metrc. Additionally, create several facility-type specific lists for updating
+        information relevant only to that specific type (i.e. sales transactions for dispensaries).
+        '''
+        print('Initializing facility list...')
+        ep = self.endpoints['facilities'][0]
+        url = f"{API_BASE_URL}{ep}"
+        self.facility_data = self.api.get_from_url_list([url])[0]
+
+        self.facilities = [facility['License']['Number'] for facility in self.facility_data]
+
+        # List all dispensaries using the boolean property in the facility data, CanSellToPatients.
+        self.dispensaries = [facility['License']['Number'] for facility in self.facility_data
+                             if facility['FacilityType']['CanSellToPatients']]
+
+        # List all manufacturing licenses using the boolean property in the facility data, CanInfuseProducts.
+        self.mipps = [facility['License']['Number'] for facility in self.facility_data
+                      if facility['FacilityType']['CanInfuseProducts']]
+
+        # List all provider licenses using the boolean property in the facility data, CanGrowPlants.
+        self.providers = [facility['License']['Number'] for facility in self.facility_data
+                          if facility['FacilityType']['CanGrowPlants']]
+
     def get_user_id(self):
         '''
         Return the fist seven (7) digits of the user API key. This can be used for creating unique database files,
@@ -88,6 +114,7 @@ class Metrc:
         Returns
         -------
         str
+            Facility start date.
         '''
         try:
             assert(facility in self.facilities)
@@ -103,36 +130,26 @@ class Metrc:
 
         return start_date
 
-    def _get_from_urls(self, urls):
+    def get_24_hour_periods(self, start_date, end_date):
         '''
-        Take a list of urls and pass them to the APICaller, getting back a list of responses in dictionaries. To be
-        precise, we get back a list of lists of dictionaries, i.e.:
-
-          |--------- 24 hr period ---------|  |--------- 24 hr period ---------|  |- 24 hr period -|
-        [ [ { package: 1 }, { package: 2 } ], [ { package: 3 }, { package: 4 } ], [ { package: 5 } ] ]
-
-        The above example assumes we have made a request from the packages endpoint for packages last modified within
-        three (3) 24 hour periods. Since we get the JSON data back converted to a Python dictionary object, and we must
-        make a separate call for each 24 hour period, we get a list of lists of dictionaries.
-
-        As many requests can be required, either for time-based queries or for ID-based queries such as when first
-        initializing your database, this is done concurrently using AIOHTTP.
+        Using a start and end datetime string in ISO 8601 format, return a list of tuples each representing no greater
+        than 24 hours. The total of the periods returned will span the range provided. Metrc requires that, for time-
+        based queries, each request covers a maximum of 24 hours. The end dates returned in the tuple are millisecond
+        resolution -- we subtract one (1) millisecond to avoid overlap of an end date and the following start date and
+        thus eliminate edge-case duplicate data. Timezone offsets can be different.
 
         Parameters
         ----------
-        urls: list[str]
-            URLs from which to make get requests.
+        start_date: str
+            Start date in ISO 8601 format (e.g. 2021-10-01T21:49:00+06:00)
+        end_date: str
+            End date in ISO 8601 format (e.g. 2021-10-17T05:19:32+06:00)
 
         Returns
         -------
-        list[list[dict]]
-            The outer list contains all individual endpoint queries. The inner list contains all the entities returned
-            an individual endpoint query. The dict is the dictionary representing the JSON object returned.
+        list[tuple(str, str)]
+            Each tuple contains a datetime range that does not exceed 24 hours as strings in ISO 8601 format.
         '''
-        response_json = self.api.get_from_url_list(urls)
-        return response_json
-
-    def get_24_hour_periods(self, start_date, end_date):
         try:
             # Verify the strings passed are ISO 8601 format and that the end date is greater than the start date.
             start_dt = datetime.fromisoformat(start_date)
@@ -164,132 +181,87 @@ class Metrc:
 
         return period_tuples
 
-    # Make a request for the users available facilities.
-    def init_facilities(self):
-        print('Initializing facility list...')
-        ep = self.endpoints['facilities'][0]
-        url = f"{API_BASE_URL}{ep}"
-        self.facility_data = self.api.get_from_url_list([url])[0]
+    def get_last_modified(self, endpoint, license_number, last_modified_start, last_modified_end, **kwargs):
+        '''
+        Retreive entities from an endpoint using last modified query parameters and flatten the list of lists of dicts.
+        Return a straight list of entities represented as dictionary objects (e.g. list[package1, package2, package3]).
 
-        self.facilities = [facility['License']['Number'] for facility in self.facility_data]
-
-        # List all dispensaries using the boolean property in the facility data, CanSellToPatients.
-        self.dispensaries = [facility['License']['Number'] for facility in self.facility_data
-                             if facility['FacilityType']['CanSellToPatients']]
-
-        # List all manufacturing licenses using the boolean property in the facility data, CanInfuseProducts.
-        self.mipps = [facility['License']['Number'] for facility in self.facility_data
-                      if facility['FacilityType']['CanInfuseProducts']]
-
-    # This method is a catch all for making queries. I know this is a clunky approach. This will change with more
-    # experience and better knowledge of design priciples and best-practices.
-    def get_data(self, endpoint, license_number, start_date, end_date, **kwargs):
+        Parameters
+        ----------
+        endpoint : str
+            Endpoint/category of endpoints defined in the class attribute, endpoints (dict). For example, "packages"
+            includes both active and inactive endpoints.
+        license_number : str
+            Facility license number to insert in query. User must have appropriate permissions for this facility granted
+            in Metrc to obain valid data.
+        last_modified_start : str
+            Beginning of time range for the overall query in ISO 8601 format.
+        last_modified_end : str
+            End of time range for the overall query in ISO 8601 format.
+        flatten : bool, optional
+            Optional keyword argument for flattening any subdictionaries in the retrieved entities.
+        '''
         try:
             assert(license_number in self.facilities)
         except AssertionError:
             print('Invalid license: User must have access to facility.')
 
+        self._print_get_message(endpoint, license_number, last_modified_start, last_modified_end)
+
         if endpoint == 'sales_transactions':
             return self.get_sales_transactions(license_number=license_number,
-                                               sales_date_start=start_date,
-                                               sales_date_end=end_date,
-                                               receipt_ids=kwargs.get('receipt_ids'),
-                                               flatten=kwargs.get('flatten'))
-        elif endpoint == 'packages':
-            return self.get_packages(license_number=license_number,
-                                     last_modified_start=start_date,
-                                     last_modified_end=end_date,
-                                     flatten=kwargs.get('flatten'))
-        elif endpoint == 'harvests':
-            return self.get_harvests(license_number=license_number,
-                                     last_modified_start=start_date,
-                                     last_modified_end=end_date)
-
-    def _generate_urls(self, endpoint, license_number, last_modified_start, last_modified_end):
-        date_ranges = self.get_24_hour_periods(last_modified_start, last_modified_end)
-        urls = []
-
-        for ep in self.endpoints[endpoint]:
-            for period in date_ranges:
-                query_dict = {'licenseNumber': license_number,
-                              'lastModifiedStart': period[0],
-                              'lastModifiedEnd': period[1]}
-                query = urllib.parse.urlencode(query_dict)
-                urls.append(f"{API_BASE_URL}{ep}?{query}")
-
-        return urls
-
-    # Return a list of packages as dictionaries.
-    def get_harvests(self, license_number, last_modified_start, last_modified_end):
-        endpoint = 'harvests'
-        self._print_get_message(endpoint, license_number, last_modified_start, last_modified_end)
-
-        urls = self._generate_urls(endpoint, license_number, last_modified_start, last_modified_end)
+                                               sales_date_start=last_modified_start,
+                                               sales_date_end=last_modified_end,
+                                               **kwargs)
 
+        urls = self._generate_last_modified_urls(endpoint, license_number, last_modified_start, last_modified_end)
         response = self._get_from_urls(urls)
-        harvests = [harvest for period in response if period for harvest in period]
-        return harvests
+        data = [entity for period in response if period for entity in period]
 
-    # Return a list of packages as dictionaries.
-    def get_packages(self, license_number, last_modified_start, last_modified_end, flatten=True):
-        endpoint = 'packages'
-        self._print_get_message(endpoint, license_number, last_modified_start, last_modified_end)
-
-        urls = self._generate_urls(endpoint, license_number, last_modified_start, last_modified_end)
-
-        response = self._get_from_urls(urls)
-        packages = [package for period in response if period for package in period]
-
-        if flatten:
-            print('Flattening packages...')
-            flattened_packages = []
-
-            for package in packages:
-                item = package.pop('Item')
-                new_item = {}
-
-                for key, value in item.items():
-                    key = 'Item' + key
-                    new_item.update({key: value})
-
-                package.update(new_item)
-                flattened_packages.append(package)
-
-            return flattened_packages
+        if kwargs.get('flatten'):
+            return self._flatten_subdictionaries(data)
         else:
-            return packages
+            return data
 
-    # Return a tuple containing a list of receipt IDs and list of receipts as dictionaries.
-    def get_sales_receipts(self, license_number, sales_date_start, sales_date_end):
-        endpoint = 'sales_receipts'
-        self._print_get_message(endpoint, license_number, sales_date_start, sales_date_end)
-
-        date_ranges = self.get_24_hour_periods(sales_date_start, sales_date_end)
-        urls = []
-
-        for ep in self.endpoints['sales_receipts']:
-            for period in date_ranges:
-                query_dict = {'licenseNumber': license_number, 'salesDateStart': period[0], 'salesDateEnd': period[1]}
-                query = urllib.parse.urlencode(query_dict)
-                urls.append(f"{API_BASE_URL}{ep}?{query}")
-
-        response = self._get_from_urls(urls)
-
-        # We get a list of lists from the async url request -- a list of receipts for each period requested, all
-        # within the top-level list. Here we flatten this and extract a list of receipt IDs for easy integration with
-        # other endpoints (transactions).
-        receipts = [receipt for period in response if period for receipt in period]
-        ids = [receipt['Id'] for receipt in receipts if isinstance(receipt, dict)]
-
-        return (ids, receipts)
-
-    # Return a list of transaction information using a list of receipt IDs for lookup. If flatten is true, return
-    # a list of elements containing transaction information per individual package. Multiple elements
-    # (transactions) can share a receipt ID. If flatten is false, return a list of receipts containing a nested list
-    # of individual packages involved. First, if no IDs are provided, get them automatically for the date range
-    # provided. This requires calling get_sales_reciepts().
     def get_sales_transactions(self, license_number, sales_date_start=None, sales_date_end=None, receipt_ids=None,
-                               flatten=True):
+                               flatten=False):
+        '''
+        Get a list of transactions for a provided list of receipt IDs, or use a sales datetime range to first trigger a
+        sales receipt lookup. This is an ID based query, so we must first get the list of valid IDs from the time-based
+        endpoint, sales_receipts. If we do not provide a list of receipt IDs, the sales date start/end is required and
+        vice versa.
+
+        Note: At the time of this writing, the GET sales receipt endpoint is broken returning a status 500 when
+        attempting to query by last modified start/end, or otherwise ignores the query parameters and defaults to the
+        last 24 hours.
+
+        Parameters
+        ----------
+        license_number : str
+            Facility license number to query. User must have appropriate permissions for this facility granted in Metrc
+            to obain valid data.
+        sales_date_start : str, optional with IDs
+            The start of the time range to query.
+        sales_date_end : str, optional with IDs
+            The end of the time range to query.
+        receipt_ids : list[str], optional with sales date start/end
+            IDs to lookup for transaction information.
+        flatten : bool, default=False
+            By default, transactions are identical to the recipt objects provided by the sales receipt endpoints except
+            the transaction field contains a populated list of individual packages that were transacted from. If
+            flattened, use the objects representing each single affected package and expand it to contain relevant
+            information for the enclosing receipt ID.
+
+        Returns
+        -------
+        list[dict]
+            A list of receipt objects represented as dictionaries.
+        '''
+        try:
+            assert(license_number in self.facilities)
+        except AssertionError:
+            print('Invalid license: User must have access to facility.')
+
         if not receipt_ids:
             try:
                 assert(sales_date_start)
@@ -298,7 +270,8 @@ class Metrc:
                 print('No receipt IDs provided: start and end date required.')
 
             print('No IDs provided when getting transactions...')
-            receipt_ids = self.get_sales_receipts(license_number, sales_date_start, sales_date_end)[0]
+            receipts = self.get_last_modified('sales_receipts', license_number, sales_date_start, sales_date_end)
+            receipt_ids = [receipt['Id'] for receipt in receipts if isinstance(receipt, dict)]
             if not receipt_ids:
                 print("No receipts in time period specified. Skipping transaction lookup.")
                 return []
@@ -347,6 +320,120 @@ class Metrc:
     def _print_get_message(self, endpoint, license_number, start_date, end_date):
         print(f'Getting {endpoint} for {license_number} between {start_date} and {end_date}')
 
+    def _flatten_subdictionaries(self, entities):
+        '''
+        Process a list of entities, such as packages, and flatten any sub-dictionaries contained within. If a
+        sub-dictionary is found, it is removed and we concatenate the name of the top-level key with the sub-key and
+        assign it the sub-value. The entity is updated with the newly flattened key-value pairs.
+
+        e.g. {'A': 'string', 'B': {'1': 'x', '2': 'y', '3': 'z'}} -> {'A': 'string', 'B1': 'x', 'B2': 'y', 'B3': 'z'}
+
+        Parameters
+        ----------
+        entities : list[dict]
+            List of entities represented as dictionaries.
+
+        Returns
+        -------
+        list[dict]
+            List of entites represented as dictionaries with sub-dictionary values flattened.
+        '''
+        print('Flattening objects...')
+        flattened_entities = []
+
+        for entity in entities:
+            keys_to_flatten = []
+            for key, value in entity.items():
+                if isinstance(value, dict):
+                    keys_to_flatten.append(key)
+
+            for key in keys_to_flatten:
+                value = entity.pop(key)
+                new_value = {}
+
+                for subkey, subvalue in value.items():
+                    cat_key = key + subkey
+                    new_value.update({cat_key: subvalue})
+
+                entity.update(new_value)
+            flattened_entities.append(entity)
+
+        return flattened_entities
+
+    def _get_from_urls(self, urls):
+        '''
+        Take a list of urls and pass them to the APICaller, getting back a list of responses in dictionaries. To be
+        precise, we get back a list of lists of dictionaries, i.e.:
+
+          |--------- 24 hr period ---------|  |--------- 24 hr period ---------|  |- 24 hr period -|
+        [ [ { package: 1 }, { package: 2 } ], [ { package: 3 }, { package: 4 } ], [ { package: 5 } ] ]
+
+        The above example assumes we have made a request from the packages endpoint for packages last modified within
+        three (3) 24 hour periods. Since we get the JSON data back converted to a Python dictionary object, and we must
+        make a separate call for each 24 hour period, we get a list of lists of dictionaries.
+
+        As many requests can be required, either for time-based queries or for ID-based queries such as when first
+        initializing your database, this is done concurrently using AIOHTTP.
+
+        Parameters
+        ----------
+        urls: list[str]
+            URLs from which to make get requests.
+
+        Returns
+        -------
+        list[list[dict]]
+            The outer list contains all individual endpoint queries. The inner list contains all the entities returned
+            an individual endpoint query. The dict is the dictionary representing the JSON object returned.
+        '''
+        response_json = self.api.get_from_url_list(urls)
+        return response_json
+
+    def _generate_last_modified_urls(self, endpoint, license_number, last_modified_start, last_modified_end):
+        '''
+        Create a list of URLs for endpoint accepting queries based on last-modified datetimes (e.g. packages, harvests).
+
+        Parameters
+        ----------
+        endpoint : str
+            Endpoint/category of endpoints defined in the class dict attribute, endpoints. For example, "packages"
+            includes both active and inactive endpoints.
+        license_number : str
+            Facility license number to insert in query. User must have appropriate permissions for this facility
+            granted in Metrc to obain valid data.
+        last_modified_start : str
+            Beginning of time range for the overall query in ISO 8601 format.
+        last_modified_end : str
+            End of time range for the overall query in ISO 8601 format.
+
+        Returns
+        -------
+        list[str]
+            A list of crafted URLs to be used for querying the Metrc API.
+        '''
+        date_ranges = self.get_24_hour_periods(last_modified_start, last_modified_end)
+        urls = []
+
+        # Note: At the time of this writing, the GET sales receipt endpoint is broken returning a status 500 when
+        # attempting to query by last modified start/end, or otherwise ignores the query parameters and defaults to the
+        # last 24 hours.
+        if endpoint == 'sales_receipts':
+            query_start_term = 'salesDateStart'
+            query_end_term = 'salesDateEnd'
+        else:
+            query_start_term = 'lastModifiedStart'
+            query_end_term = 'lastModifiedEnd'
+
+        for ep in self.endpoints[endpoint]:
+            for period in date_ranges:
+                query_dict = {'licenseNumber': license_number,
+                              query_start_term: period[0],
+                              query_end_term: period[1]}
+                query = urllib.parse.urlencode(query_dict)
+                urls.append(f"{API_BASE_URL}{ep}?{query}")
+
+        return urls
+
 
 if __name__ == '__main__':
     import os
